//  dibapi.cpp    functions for device independent bitmaps
//  portions copyright microsoft, portions dovermyer
//

#include "stdafx.h" 
#include "afxext.h"
#include "dibapi.h"
#include <dos.h>

#define DIB_HEADER_MARKER   ((WORD) ('M' << 8) | 'B')

//********************************************************************
BOOL WINAPI PaintDIB(HDC     hDC,
					LPRECT  lpDCRect,
					HDIB    hDIB,
					LPRECT  lpDIBRect,
					CPalette* pPal)
{
	LPSTR    lpDIBHdr;            // Pointer to BITMAPINFOHEADER
	LPSTR    lpDIBBits;           // Pointer to DIB bits
	BOOL     bSuccess=FALSE;      // Success/fail flag
	HPALETTE hPal=NULL;           // Our DIB's palette
	HPALETTE hOldPal=NULL;        // Previous palette

	/* Check for valid DIB handle */
	if (hDIB == NULL)
		return FALSE;

	/* Lock down the DIB, and get a pointer to the beginning of the bit
	 *  buffer
	 */
	lpDIBHdr  = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);
	lpDIBBits = ::FindDIBBits(lpDIBHdr);

	// Get the DIB's palette, then select it into DC
	if (pPal != NULL)
	{
		hPal = (HPALETTE) pPal->m_hObject;

		// Select as background since we have
		// already realized in forground if needed
		hOldPal = ::SelectPalette(hDC, hPal, TRUE);
	}

	/* Make sure to use the stretching mode best for color pictures */
	::SetStretchBltMode(hDC, STRETCH_DELETESCANS);

	/* Determine whether to call StretchDIBits() or SetDIBitsToDevice() */
	if ((RECTWIDTH(lpDCRect)  == RECTWIDTH(lpDIBRect)) &&
	   (RECTHEIGHT(lpDCRect) == RECTHEIGHT(lpDIBRect)))
		bSuccess = ::SetDIBitsToDevice(hDC,                    // hDC
								   lpDCRect->left,             // DestX
								   lpDCRect->top,              // DestY
								   RECTWIDTH(lpDCRect),        // nDestWidth
								   RECTHEIGHT(lpDCRect),       // nDestHeight
								   lpDIBRect->left,            // SrcX
								   (int)DIBHeight(lpDIBHdr) -
									  lpDIBRect->top -
									  RECTHEIGHT(lpDIBRect),   // SrcY
								   0,                          // nStartScan
								   (WORD)DIBHeight(lpDIBHdr),  // nNumScans
								   lpDIBBits,                  // lpBits
								   (LPBITMAPINFO)lpDIBHdr,     // lpBitsInfo
								   DIB_RGB_COLORS);            // wUsage
   else
	  bSuccess = ::StretchDIBits(hDC,                          // hDC
							   lpDCRect->left,                 // DestX
							   lpDCRect->top,                  // DestY
							   RECTWIDTH(lpDCRect),            // nDestWidth
							   RECTHEIGHT(lpDCRect),           // nDestHeight
							   lpDIBRect->left,                // SrcX
							   lpDIBRect->top,                 // SrcY
							   RECTWIDTH(lpDIBRect),           // wSrcWidth
							   RECTHEIGHT(lpDIBRect),          // wSrcHeight
							   lpDIBBits,                      // lpBits
							   (LPBITMAPINFO)lpDIBHdr,         // lpBitsInfo
							   DIB_RGB_COLORS,                 // wUsage
							   SRCCOPY);                       // dwROP

   ::GlobalUnlock((HGLOBAL) hDIB);

	/* Reselect old palette */
	if (hOldPal != NULL)
	{
		::SelectPalette(hDC, hOldPal, TRUE);
	}

   return bSuccess;
}

//********************************************************************
BOOL WINAPI CreateDIBPalette(HDIB hDIB, CPalette* pPal)
{
	LPLOGPALETTE lpPal;      // pointer to a logical palette
	HANDLE hLogPal;          // handle to a logical palette
	HPALETTE hPal = NULL;    // handle to a palette
	int i;                   // loop index
	WORD wNumColors;         // number of colors in color table
	LPSTR lpbi;              // pointer to packed-DIB
	LPBITMAPINFO lpbmi;      // pointer to BITMAPINFO structure (Win3.0)
	LPBITMAPCOREINFO lpbmc;  // pointer to BITMAPCOREINFO structure (old)
	BOOL bWinStyleDIB;       // flag which signifies whether this is a Win3.0 DIB
	BOOL bResult = FALSE;

	/* if handle to DIB is invalid, return FALSE */

	if (hDIB == NULL) return FALSE;

   lpbi = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);

   /* get pointer to BITMAPINFO (Win 3.0) */
   lpbmi = (LPBITMAPINFO)lpbi;

   /* get pointer to BITMAPCOREINFO (old 1.x) */
   lpbmc = (LPBITMAPCOREINFO)lpbi;

   /* get the number of colors in the DIB */
   wNumColors = ::DIBNumColors(lpbi);

   if (wNumColors != 0)
   	{
		/* allocate memory block for logical palette */
		hLogPal = ::GlobalAlloc(GHND, sizeof(LOGPALETTE)
									+ sizeof(PALETTEENTRY)
									* wNumColors);

		/* if not enough memory, clean up and return NULL */
		if (hLogPal == 0)
			{
			::GlobalUnlock((HGLOBAL) hDIB);
			return FALSE;
			}

		lpPal = (LPLOGPALETTE) ::GlobalLock((HGLOBAL) hLogPal);

		/* set version and number of palette entries */
		lpPal->palVersion = PALVERSION;
		lpPal->palNumEntries = (WORD)wNumColors;

		/* is this a Win 3.0 DIB? */
		bWinStyleDIB = IS_WIN30_DIB(lpbi);
		for (i = 0; i < (int)wNumColors; i++)
			{
			if (bWinStyleDIB)
				{
				lpPal->palPalEntry[i].peRed = lpbmi->bmiColors[i].rgbRed;
				lpPal->palPalEntry[i].peGreen = lpbmi->bmiColors[i].rgbGreen;
				lpPal->palPalEntry[i].peBlue = lpbmi->bmiColors[i].rgbBlue;
				lpPal->palPalEntry[i].peFlags = 0;
				}
			else
				{
				lpPal->palPalEntry[i].peRed = lpbmc->bmciColors[i].rgbtRed;
				lpPal->palPalEntry[i].peGreen = lpbmc->bmciColors[i].rgbtGreen;
				lpPal->palPalEntry[i].peBlue = lpbmc->bmciColors[i].rgbtBlue;
				lpPal->palPalEntry[i].peFlags = 0;
				}
			}

		/* create the palette and get handle to it */
		bResult = pPal->CreatePalette(lpPal);
		::GlobalUnlock((HGLOBAL) hLogPal);
		::GlobalFree((HGLOBAL) hLogPal); 
		
		} 
		                                   
	if (!bResult) 
		bResult = pPal->CreateStockObject(DEFAULT_PALETTE);
	::GlobalUnlock((HGLOBAL) hDIB);

	return bResult;
}

//********************************************************************
LPSTR WINAPI FindDIBBits(LPSTR lpbi)
{
	return (lpbi + *(LPDWORD)lpbi + ::PaletteSize(lpbi));
}


//********************************************************************
DWORD WINAPI DIBWidth(LPSTR lpDIB)
{
	LPBITMAPINFOHEADER lpbmi;  // pointer to a Win 3.0-style DIB
	LPBITMAPCOREHEADER lpbmc;  // pointer to an other-style DIB

	/* point to the header (whether Win 3.0 and old) */

	lpbmi = (LPBITMAPINFOHEADER)lpDIB;
	lpbmc = (LPBITMAPCOREHEADER)lpDIB;

	/* return the DIB width if it is a Win 3.0 DIB */
	if (IS_WIN30_DIB(lpDIB))
		return lpbmi->biWidth;
	else  /* it is an other-style DIB, so return its width */
		return (DWORD)lpbmc->bcWidth;
}


//********************************************************************
DWORD WINAPI DIBHeight(LPSTR lpDIB)
{
	LPBITMAPINFOHEADER lpbmi;  // pointer to a Win 3.0-style DIB
	LPBITMAPCOREHEADER lpbmc;  // pointer to an other-style DIB

	/* point to the header (whether old or Win 3.0 */

	lpbmi = (LPBITMAPINFOHEADER)lpDIB;
	lpbmc = (LPBITMAPCOREHEADER)lpDIB;

	/* return the DIB height if it is a Win 3.0 DIB */
	if (IS_WIN30_DIB(lpDIB))
		return lpbmi->biHeight;
	else  /* it is an other-style DIB, so return its height */
		return (DWORD)lpbmc->bcHeight;
}


//********************************************************************
WORD WINAPI PaletteSize(LPSTR lpbi)
{
   /* calculate the size required by the palette */
   if (IS_WIN30_DIB (lpbi))
	  return (WORD)(::DIBNumColors(lpbi) * sizeof(RGBQUAD));
   else
	  return (WORD)(::DIBNumColors(lpbi) * sizeof(RGBTRIPLE));
}

//********************************************************************
WORD WINAPI DIBNumColors(LPSTR lpbi)
{
	WORD wBitCount;  // DIB bit count

	/*  If this is a Windows-style DIB, the number of colors in the
	 *  color table can be less than the number of bits per pixel
	 *  allows for (i.e. lpbi->biClrUsed can be set to some value).
	 *  If this is the case, return the appropriate value.
	 */

	if (IS_WIN30_DIB(lpbi))
		{
		DWORD dwClrUsed;

		dwClrUsed = ((LPBITMAPINFOHEADER)lpbi)->biClrUsed;
		if (dwClrUsed != 0)
			return (WORD)dwClrUsed;
		}

	/*  Calculate the number of colors in the color table based on
	 *  the number of bits per pixel for the DIB.
	 */
	if (IS_WIN30_DIB(lpbi))
		wBitCount = ((LPBITMAPINFOHEADER)lpbi)->biBitCount;
	else
		wBitCount = ((LPBITMAPCOREHEADER)lpbi)->bcBitCount;

	/* return number of colors based on bits per pixel */
	switch (wBitCount)
		{
		case 1:
			return 2;

		case 4:
			return 16;

		case 8:
			return 256;

		default:
			return 0;
		}
}


//////////////////////////////////////////////////////////////////////////
//// Clipboard support

//********************************************************************
HANDLE WINAPI CopyHandle (HANDLE h)
{
	BYTE huge *lpCopy;
	BYTE huge *lp;
	HANDLE     hCopy;
	DWORD      dwLen;

	if (h == NULL)
		return NULL;

	dwLen = ::GlobalSize((HGLOBAL) h);

	if ((hCopy = (HANDLE) ::GlobalAlloc (GHND, dwLen)) != NULL)
		{
		lpCopy = (BYTE huge *) ::GlobalLock((HGLOBAL) hCopy);
		lp     = (BYTE huge *) ::GlobalLock((HGLOBAL) h);

		while (dwLen--)
			*lpCopy++ = *lp++;

		::GlobalUnlock((HGLOBAL) hCopy);
		::GlobalUnlock((HGLOBAL) h);
		}

	return hCopy;
}

//******************************************************************** 
// hD			:handle to DIB - this is the input
// pB     :pointer to output bitmap
// pW     :pointer to host CWnd
// pP     :pointer to a CPalette dynamically alloc'd in this function - goes with the bitmap
// sr     :sizing rect - size output to fit within this rect
BOOL WINAPI	DIBtoDDB(HGLOBAL hD,CBitmap* pB,CWnd* pW,CPalette* pP,CRect sr)
{
  LPSTR	 			bi, bits;
  CPalette*		poP;
  WORD				x,y,maxxy;
  CRect				rc;
  CBitmap*		poB;   
  CDC*				pDC;
  CDC					memDC;            
  
  if (hD == NULL) return FALSE;    			//nothing in, nothing out
  bi=(LPSTR)::GlobalLock((HGLOBAL)hD);  //get a pointer to memory block
  if (bi == NULL) return FALSE;
  
  ::CreateDIBPalette((HDIB)hD, pP);     //build palette from DIB color table
  
  pDC=pW->GetDC();                      //get a device context from host window
  memDC.CreateCompatibleDC(pDC);        //get compatible memory dc
  
  x=(WORD)::DIBWidth(bi);               //adjust rc for DIB aspect ratio
  y=(WORD)::DIBHeight(bi);              //rc now represents the target width/height
  maxxy = max(x,y);       
  ::SetRect(rc,0,0,::MulDiv(sr.right,x,maxxy), ::MulDiv(sr.bottom,y,maxxy));
  
  pB->CreateCompatibleBitmap(pDC,rc.right,rc.bottom);  //create output DDB
  poB = memDC.SelectObject(pB);                        //setup memDC 
  if (pP != NULL) poP = memDC.SelectPalette(pP,FALSE);
  memDC.RealizePalette(); 
  
  bits=FindDIBBits(bi);                                //get pointer to bits
  memDC.SetStretchBltMode(STRETCH_DELETESCANS);        //set stretch mode for color
  ::StretchDIBits(memDC.m_hDC,0,0,rc.right,rc.bottom,0,0,x,y,bits,
  		(LPBITMAPINFO)bi,DIB_RGB_COLORS,SRCCOPY);        //scale DIB onto DDB
  		
  ::GlobalUnlock((HGLOBAL)hD);                         //unlock DIB
  if (poP != NULL) memDC.SelectPalette(poP,TRUE);      //restore old palette
  memDC.SelectObject(poB);                             //deselect DDB
  pW->ReleaseDC(pDC);                                  //release DC
 	return TRUE;                                         // pB contains the desired DDB, pP the matching palette
}


//********************************************************************

BOOL WINAPI SaveDIB(HDIB hDib, CFile& file)
{
	BITMAPFILEHEADER bmfHdr; // Header for Bitmap file
	LPBITMAPINFOHEADER lpBI;   // Pointer to DIB info structure
	DWORD dwDIBSize;

	if (hDib == NULL)
		return FALSE;

	/*
	 * Get a pointer to the DIB memory, the first of which contains
	 * a BITMAPINFO structure
	 */
	lpBI = (LPBITMAPINFOHEADER) ::GlobalLock((HGLOBAL) hDib);
	if (lpBI == NULL)
		return FALSE;

	if (!IS_WIN30_DIB(lpBI))
		{
		::GlobalUnlock((HGLOBAL) hDib);
		return FALSE;       // It's an other-style DIB (save not supported)
		}

	/*
	 * Fill in the fields of the file header
	 */

	/* Fill in file type (first 2 bytes must be "BM" for a bitmap) */
	bmfHdr.bfType = DIB_HEADER_MARKER;  // "BM"

	// Calculating the size of the DIB is a bit tricky (if we want to
	// do it right).  The easiest way to do this is to call GlobalSize()
	// on our global handle, but since the size of our global memory may have
	// been padded a few bytes, we may end up writing out a few too
	// many bytes to the file (which may cause problems with some apps).
	//
	// So, instead let's calculate the size manually (if we can)
	//
	// First, find size of header plus size of color table.  Since the
	// first DWORD in both BITMAPINFOHEADER and BITMAPCOREHEADER conains
	// the size of the structure, let's use this.

	dwDIBSize = *(LPDWORD)lpBI + ::PaletteSize((LPSTR)lpBI);  // Partial Calculation

	// Now calculate the size of the image

	if ((lpBI->biCompression == BI_RLE8) || (lpBI->biCompression == BI_RLE4))
		{
		// It's an RLE bitmap, we can't calculate size, so trust the
		// biSizeImage field

		dwDIBSize += lpBI->biSizeImage;
		}
	else
		{
		DWORD dwBmBitsSize;  // Size of Bitmap Bits only

		// It's not RLE, so size is Width (DWORD aligned) * Height

		dwBmBitsSize = WIDTHBYTES((lpBI->biWidth)*((DWORD)lpBI->biBitCount)) * lpBI->biHeight;

		dwDIBSize += dwBmBitsSize;

		// Now, since we have calculated the correct size, why don't we
		// fill in the biSizeImage field (this will fix any .BMP files which
		// have this field incorrect).

		lpBI->biSizeImage = dwBmBitsSize;
		}


	// Calculate the file size by adding the DIB size to sizeof(BITMAPFILEHEADER)

	bmfHdr.bfSize = dwDIBSize + sizeof(BITMAPFILEHEADER);
	bmfHdr.bfReserved1 = 0;
	bmfHdr.bfReserved2 = 0;

								//calc offset of bits
	bmfHdr.bfOffBits = (DWORD)sizeof(BITMAPFILEHEADER) + lpBI->biSize
											  + PaletteSize((LPSTR)lpBI);

	TRY
		{
		// Write the file header
		file.Write((LPSTR)&bmfHdr, sizeof(BITMAPFILEHEADER));
		//
		// Write the DIB header and the bits
		//
		file.WriteHuge(lpBI, dwDIBSize);
		}
	CATCH (CFileException, e)
		{
		::GlobalUnlock((HGLOBAL) hDib);
		THROW_LAST();
		}
	END_CATCH

	::GlobalUnlock((HGLOBAL) hDib);
	return TRUE;
}

//********************************************************************
// create a packed DIB from disk file
HDIB WINAPI ReadDIBFile(CFile& file)
{
	BITMAPFILEHEADER 	bmfHeader;
	DWORD 						dwBitsSize,dwFileSize;
	HDIB 							hDIB;
	LPSTR 						pDIB;
	BYTE huge * 			hpDIB; //hugh pointer into packed DIB memory block
	DWORD 						actual;
  UINT 							ctSize,biSize;

	dwFileSize = file.GetLength();

	 				// read the DIB file header and check if it's valid.
	if ((file.Read((LPSTR)&bmfHeader, sizeof(bmfHeader)) !=
			sizeof(bmfHeader)) || (bmfHeader.bfType != DIB_HEADER_MARKER))
		return NULL;
                         
	dwBitsSize=dwFileSize-bmfHeader.bfOffBits;

				// Allocate memory for DIB
	hDIB = (HDIB) ::GlobalAlloc(GHND,dwFileSize-sizeof(BITMAPFILEHEADER));
	if (hDIB == 0) return NULL;

	pDIB = (LPSTR) ::GlobalLock((HGLOBAL) hDIB); 
				//point hpDIB to global block
  hpDIB = (BYTE huge *)pDIB;
				//read the header size - sanity check
	actual=file.Read(hpDIB,sizeof(DWORD));
	hpDIB +=actual;  // update hugh pointer 
	biSize=(UINT)((BITMAPINFOHEADER far *)pDIB)->biSize;
	if ( (biSize != sizeof(BITMAPINFOHEADER)) && 
	     (biSize != sizeof(BITMAPCOREHEADER)) )	
		{
		::GlobalUnlock((HGLOBAL) hDIB);
		::GlobalFree((HGLOBAL) hDIB);
		return NULL;
		}
                          
				//read the rest of header  
	UINT rest= biSize - sizeof(DWORD);
	actual=file.Read(hpDIB,rest);
	hpDIB +=actual;  // update hugh pointer
	  
				//read the color table  into memory
	ctSize=(biSize == sizeof(BITMAPINFOHEADER))?
			sizeof(RGBQUAD)*DIBNumColors((LPSTR)pDIB):
			sizeof(RGBTRIPLE)*DIBNumColors((LPSTR)pDIB);
	actual=file.Read(hpDIB,ctSize);
	hpDIB +=actual;  //update hugh pointer
	
				//find the bits
	file.Seek(bmfHeader.bfOffBits,CFile::begin); 
				//read the dib bits
	actual=file.ReadHuge(hpDIB,dwBitsSize);
	hpDIB +=actual;  //update hugh pointer
	
        //tidy up
	::GlobalUnlock((HGLOBAL) hDIB);
	return hDIB;
}

//******************************************************************** 
// open DIB file, call ReadDIBFile()
HDIB WINAPI LoadDIBFile(PSTR fname, DWORD ofsstart)
{
	CFile file;
	CFileException fe;
	HDIB m_hDIB ; 
	
	if (!file.Open(fname, CFile::modeRead | CFile::shareDenyWrite, &fe))
		return 0;

	m_hDIB = ::ReadDIBFile(file);
	return m_hDIB;
} 


//********************************************************************
//create a DIB from a CBitmap object
HGLOBAL	  WINAPI  DDBtoDIB(CBitmap* pBmp, CPalette* pPal,CDC* pDC)
{
	CPalette*							poldPal;
	BITMAP								bm;
	HDIB            			hDIB=0; 
	LPSTR									pDIB=NULL;
	LPSTR           			pBits=NULL;
	int										iRes=0,iColor=0;  
	DWORD           			dwSize=0; 
	BITMAPINFOHEADER far* pBI;
	                   //realize the palette if available
	if (pPal)          
		poldPal=pDC->SelectPalette(pPal,FALSE);
	pDC->RealizePalette();	                   
	                   //calc some DIB parameters
  pBmp->GetObject(sizeof(BITMAP),&bm);
  iRes=bm.bmPlanes*bm.bmBitsPixel; 
                     //if < 24 bit, GDI always creates full sized color table
                     //DIB resolution set to current bitmap (ie adapter) resolution
  if (iRes < 24)     
  	iColor = 1<<iRes;
  dwSize=  CalcDIBBits(bm.bmWidth,bm.bmHeight,iRes);  
                     //carve out packed DIB memory block
  hDIB=(HDIB)::GlobalAlloc(GHND,sizeof(BITMAPINFOHEADER)+
  			iColor*sizeof(RGBQUAD)+	dwSize);        
  			             //get pointers to block
  pDIB=(LPSTR)::GlobalLock(hDIB);
  pBI=(BITMAPINFOHEADER far *)pDIB;                                            
                     //fillin header info
  pBI->biSize=sizeof(BITMAPINFOHEADER);
  pBI->biWidth=bm.bmWidth;
  pBI->biHeight=bm.bmHeight;
  pBI->biPlanes = 1;
  pBI->biBitCount = iRes;
  pBI->biCompression = BI_RGB;
  pBI->biSizeImage = dwSize;
  pBits=(LPSTR)MK_FP(_FP_SEG(pDIB),sizeof(BITMAPINFOHEADER)+iColor*sizeof(RGBQUAD));
                     //get the bits & color table from GDI
  ::GetDIBits(pDC->m_hDC,(HBITMAP)pBmp->m_hObject,
  		0,bm.bmHeight,pBits,(BITMAPINFO far*)pBI,DIB_RGB_COLORS); 
  		               //clean up
  ::GlobalUnlock(hDIB);
  if (pPal) pDC->SelectPalette(poldPal,TRUE);
  		
  return hDIB;
}   

//********************************************************************
// compute size of DIB bits given width,height,resolution(bitsperpixel)
DWORD     WINAPI  CalcDIBBits(DWORD W,DWORD H,DWORD R)
{
	return WIDTHBYTES(W*R)*H;
}

//********************************************************************                      
//get the MM_HIMETRIC extents corresponding to DIB width,height
void WINAPI GetDIBScale(HGLOBAL H,CDC* pDC, CWnd* pWnd, LONG& X, LONG& Y)
{
	int 							oldMode;
	CPoint 						xP(0,0),yP(0,0); 
	BITMAPINFOHEADER far* pBMI;

  X = Y = 0;
  if (!H) return;
                   //get a pointer to to DIB
	pBMI = (BITMAPINFOHEADER far*)::GlobalLock(H);
	if (!pBMI) return;
                   //set mapping mode to himetric
	oldMode = pDC->SetMapMode(MM_HIMETRIC);
                   //set some 'points' that represent bitmaps width,height
  xP.x = (int)pBMI->biWidth;
  yP.y = (int)pBMI->biHeight;                                             
                   //set origin to 0.0
  pDC->SetViewportOrg(0,0);           
  		             //'convert' the points from device to logical units
  pDC->DPtoLP(&xP);pDC->DPtoLP(&yP); 
                   //restore old mapping mode
  pDC->SetMapMode(oldMode);   
  ::GlobalUnlock(H);               
  	               //return the extents
  X=xP.x;Y=yP.y;
  return;    
}

//********************************************************************                                              
//get a metafilepict that contains copy of hDIB
HGLOBAL WINAPI DIBtoPICT(HGLOBAL hDIB, CDC* pDC,CWnd* pWnd)
{
	BITMAPINFO far* 			pBMI;   
	BITMAPINFOHEADER far* pBMIH;	
	LPSTR									lpBits;	  
	HMETAFILE							hMF;
	HGLOBAL								hMFP;	
	METAFILEPICT far*			pMFP;
	LONG									X,Y;
	CMetaFileDC           MetaDC;  
	CPalette							Pal,defaultPal;
	
	if (hDIB == NULL) return 0;
	              //get width,height in MM_HIMETRIC logical units
	GetDIBScale(hDIB,pDC,pWnd, X, Y);
               //get a palette from DIB
	::CreateDIBPalette((HDIB)hDIB,&Pal);
	defaultPal.CreateStockObject(DEFAULT_PALETTE);
	             //create the metafile DC
	MetaDC.Create();
	             // get pointers to the DIB
	pBMI=(BITMAPINFO far *)::GlobalLock(hDIB); 
	pBMIH=(BITMAPINFOHEADER far*)pBMI;	
	lpBits=::FindDIBBits((LPSTR)pBMI);
	             //set mapping mode in metaDC
	MetaDC.SetMapMode(MM_ANISOTROPIC);
	             //set window (logical) extents to size of DIB
	MetaDC.SetWindowExt((int)X,(int)Y);   
	             //setup the palette
	MetaDC.SelectPalette(&Pal,FALSE);
	MetaDC.RealizePalette();  
							 //set the stretch mode
	MetaDC.SetStretchBltMode(STRETCH_DELETESCANS);
	             //blt the dib to the metaDC
	StretchDIBits(MetaDC.m_hDC,0,0,(int)X,(int)Y,
			0,0,(int)pBMIH->biWidth,(int)pBMIH->biHeight,
			lpBits,pBMI,DIB_RGB_COLORS,SRCCOPY); 
			         //deselect palette	
	MetaDC.SelectPalette(&defaultPal,TRUE); 
			         //we're done recording
	hMF=MetaDC.Close(); 
	             //allocate METAFILEPICT structure
	hMFP= (HGLOBAL)::GlobalAlloc(GHND, sizeof(METAFILEPICT));
  pMFP = (METAFILEPICT far *)::GlobalLock(hMFP);
               //fill in the structure
  pMFP->hMF = hMF;   
  pMFP->xExt = (int)X;
  pMFP->yExt = (int)Y; 
  pMFP->mm = MM_ANISOTROPIC;
  GlobalUnlock(hMFP); 
               // return the pict!
  return hMFP;  
 }