// dibview.cpp : implementation of the CDibView class    
// display bitmap graphics in a view window
// doug overmyer 7/1/93
//

#include "stdafx.h"  

#include "albuminc.h"
#include "album.h"
#include "dibdoc.h"
#include "mainfrm.h"
#include "dibview.h"
#include "dibapi.h" 
#include "wmf.h"    
#include <strstrea.h>    

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////  
//utility function to draw a 3d rectangle
void Rectangle3D(CDC* pDC, CRect r, UINT style = 1);  

/////////////////////////////////////////////////////////////////////////////
// CDibView

IMPLEMENT_DYNCREATE(CDibView, CScrollView)

BEGIN_MESSAGE_MAP(CDibView, CScrollView)
	//{{AFX_MSG_MAP(CDibView)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, OnUpdateEditCopy)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_MESSAGE(WM_DOREALIZE, OnDoRealize)
	ON_WM_SIZE()
	ON_WM_NCRBUTTONUP()
	//}}AFX_MSG_MAP

	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CScrollView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CScrollView::OnFilePrintPreview)
END_MESSAGE_MAP()

///////////////////  Construction/Destruction    ////////////////////////
// CDibView construction/destruction

CDibView::CDibView()
{
}

CDibView::~CDibView()
{
}
 
BOOL CDibView::PreCreateWindow(CREATESTRUCT& cs)
{   
	WNDCLASS wc;
	::GetClassInfo(AfxGetInstanceHandle(),"AfxWnd",&wc);    
	cs.lpszClass = AfxRegisterWndClass(wc.style,
		wc.hCursor,
		(HBRUSH)::GetStockObject(LTGRAY_BRUSH),       // change the background
		AfxGetApp()->LoadIcon(IDR_DIBTYPE));          // change the icon

 	return CView::PreCreateWindow(cs);
}
//////////////////////  On Handlers     /////////////////////////////////// 


void CDibView::OnActivateView(BOOL bActivate, CView* pActivateView,
					CView* pDeactiveView)
{
	CScrollView::OnActivateView(bActivate, pActivateView, pDeactiveView);

	if (bActivate)
		{
		ASSERT(pActivateView == this);
		OnDoRealize((WPARAM)m_hWnd, 0);   // force palette to the foreground
		}
}

void CDibView::OnDraw(CDC* pDC)
{
	CDibDoc* 			pDoc = GetDocument(); 
	CMainFrame* 	pMF = GetMainFrame();	    
	CMDIChildWnd* pDocFrame = GetDocFrame();
	HDIB 					hDIB = pDoc->GetHDIB();
	BITMAP        bm;
	CRect					rc3D,rcImg,rcDest;
	int						x = 0;   
	CDC 					memDC;     
			
	if (!hDIB) return;  

	if (GetParentFrame()->IsZoomed())    //compute image rect
		{	
		LPSTR lpDIB = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);
		rcImg.top = rcImg.left = 0;
		rcImg.right = (int) ::DIBWidth(lpDIB);
		rcImg.bottom = (int) ::DIBHeight(lpDIB); 
		::GlobalUnlock((HGLOBAL) hDIB);
		}
	else
		{
		pDoc->GetDDB()->GetObject(sizeof(BITMAP),&bm);
		rcImg.SetRect(0,0,bm.bmWidth,bm.bmHeight);		
		}

	GetClientRect(rcDest);            //compute display rect
	rcDest.left = max((rcDest.right - rcImg.right ) / 2,0); 
	rcDest.right = rcDest.left + rcImg.right;
	rcDest.top = max((rcDest.bottom - rcImg.bottom)/2,0);
	rcDest.bottom = rcDest.top + rcImg.bottom;  
	
	if (pDocFrame->IsZoomed())        //if zoomed blit image
		::PaintDIB(pDC->m_hDC, &rcDest, pDoc->GetHDIB(),
				&rcImg, pDoc->GetDocPalette());
	else
		{                               //if not zoomed, frame it first
		GetClientRect(&rc3D); //compute rect for 3d frame
		x = max(-int(rc3D.Width() * .05),-int(rc3D.Height() * .05)); // margin=5% of smallest dimension
		rc3D.InflateRect(x,x); //apply deflation factor to 3d rect    
		::Rectangle3D(pDC,rc3D,1); //draw the rectangle
      //display the thumbnail
		memDC.CreateCompatibleDC(pDC);                  
		CPalette* poPal = pDC->SelectPalette(pDoc->GetDDBPalette(),TRUE); // palette really managed in OnDoRealize
		pDC->RealizePalette();
		CBitmap* poB = memDC.SelectObject(pDoc->GetDDB()); 
		pDC->BitBlt(rcDest.left,rcDest.top,
				rcDest.right-rcDest.left,rcDest.bottom-rcDest.top,&memDC,0,0,SRCCOPY);
		memDC.SelectObject(poB);        //clean up
		pDC->SelectPalette(poPal,TRUE);
		}
}  


LRESULT CDibView::OnDoRealize(WPARAM wParam, LPARAM)
{
	ASSERT(wParam != NULL);
	CDibDoc* pDoc = GetDocument();
	if (pDoc->GetHDIB() == NULL)
		return 0L;  // must be a new document

	CPalette* pPal = pDoc->GetDocPalette();
	if (pPal != NULL)
		{
		CMainFrame* pAppFrame = (CMainFrame*) AfxGetApp()->m_pMainWnd;
		ASSERT(pAppFrame->IsKindOf(RUNTIME_CLASS( CMainFrame )));

		CClientDC appDC(pAppFrame);
		// All views but one should be a background palette.
		// wParam contains a handle to the active view, so the SelectPalette
		// bForceBackground flag is FALSE only if wParam == m_hWnd (this view)
		CPalette* oldPalette = appDC.SelectPalette(pPal, ((HWND)wParam) != m_hWnd);

		if (oldPalette != NULL)
			{
			UINT nColorsChanged = appDC.RealizePalette();
			if (nColorsChanged > 0)
				pDoc->UpdateAllViews(NULL);
			appDC.SelectPalette(oldPalette, TRUE);
			}
		else
			TRACE0("\tSelectPalette failed in CDibView::OnPaletteChanged\n");
		}

	return 0L;
} 

//copy  DIB image to clipboard in DIB and metafilepict formats
void CDibView::OnEditCopy()
{
	CDibDoc* 			pDoc = GetDocument(); 
	CMainFrame* 	pMF = GetMainFrame();	    
	CMDIChildWnd* pDocFrame = GetDocFrame();
	HDIB 					hDIB = pDoc->GetHDIB();               
	CDC*					pDC;
	HGLOBAL				hMFP;

	if ( !hDIB ) return; 
	       //construct a metfile version of DIB
	hMFP = DIBtoPICT(hDIB,pDC=GetDC(),this);
	ReleaseDC(pDC);
         //copy the formats to the clipboard
	if (OpenClipboard())
		{
		BeginWaitCursor();
		EmptyClipboard();
		SetClipboardData (CF_DIB, CopyHandle((HANDLE) hDIB) );  
		SetClipboardData (CF_METAFILEPICT, hMFP );
		CloseClipboard();
		EndWaitCursor();
		}                          
	else if (hMFP)    //if failure delete the metafile copy
		::DelPICT(hMFP);
}


void CDibView::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();
	ASSERT(GetDocument() != NULL);

	SetScrollSizes(MM_TEXT, GetDocument()->GetDocSize());
}

void CDibView::OnPrepareDC(CDC* pDC,CPrintInfo* pInfo)
{
	CScrollView::OnPrepareDC(pDC,pInfo);
	if (!pInfo->m_bContinuePrinting)     // if not first page(always TRUE), use m_bIsMoreToPrint;
		pInfo->m_bContinuePrinting = m_bIsMoreToPrint;
}

BOOL CDibView::OnPreparePrinting(CPrintInfo* pInfo)
{      
	// default preparation
	return DoPreparePrinting(pInfo);
} 

// print all open DIB views 
// each view has its own document, so whatever view is active gets called and runs the whole show
// we do this by obtaining a list of all open views from the MainFrame                
// calculate some page sizes, resolution, & adjust for margins  
// print a page header
// based on the page number, loop through the view list to reach the first view to print on this page
// compute a target rect in the DC that describes where each view should render itself
// ask each view to print its caption and reduce the target rect as needed, then render self as thumbnail
// continue until page is finished
// set the m_bIsMoreToPrint flag which is used to control the process in OnPrepareDC()
void CDibView::OnPrint(CDC* pDC,CPrintInfo* pInfo)
{    
	CObList* 		pviewList; 
	POSITION		vpos;
	CDibView*		pview;
	CRect				tr;
	WORD				i = 0,w=0,h=0,r=0,c=0,pgStart=0,pgEnd=0;
	CAlbumApp* pApp = ((CAlbumApp*)AfxGetApp());   
	SState __far*	st= (SState far*)
					(pApp->m_pMainWnd->SendMessage(UM_GETSTATEINFO,0,0)); // get state info
	CPoint			grid(st->m_across,st->m_down);
	CStsMtr * pStatusBar = (CStsMtr *)
				LOWORD(pApp->m_pMainWnd->SendMessage(UM_GETSTATUSBAR,0,0)); // get pointer to status bar 

	int cxInch = pDC->GetDeviceCaps(LOGPIXELSX);
	int cyInch = pDC->GetDeviceCaps(LOGPIXELSY); 
	//compute the margins, store in a CRect
	CRect mRect(int(st->m_lMargin*cxInch),int(st->m_tMargin*cyInch),
			int(st->m_rMargin*cxInch),int(st->m_bMargin*cyInch));
	pDC->DPtoLP(mRect);//convert to logical units           
  //adjust drawing area by margins
	pInfo->m_rectDraw.left +=mRect.left;   
	pInfo->m_rectDraw.top +=mRect.top;
	pInfo->m_rectDraw.right -=mRect.right;
	pInfo->m_rectDraw.bottom -=mRect.bottom;
	  
	// get size of printer page (in pixels)
	int cxPage = pDC->GetDeviceCaps(HORZRES);
	int cyPage = pDC->GetDeviceCaps(VERTRES);
	
	pviewList = pApp->GetViewList(); 
	vpos=pviewList->GetHeadPosition();  
	
	int cxPrint=pInfo->m_rectDraw.Width();
	int cyPrint=pInfo->m_rectDraw.Height();
	w=::MulDiv((pInfo->m_rectDraw.Width()/grid.x),9,10);    //10% margin between thumbnails
	h=::MulDiv((pInfo->m_rectDraw.Height()/grid.y),9,10);  
	
	pgStart=(pInfo->m_nCurPage-1)*grid.x*grid.y;   //calc starting & ending view for this page
	pgEnd = pgStart+(grid.x*grid.y);
	m_bIsMoreToPrint=FALSE;                        // reset this later
	int uCnt = pviewList->GetCount();              // for progress meter
	
	if (!pInfo->m_bPreview)
  	pStatusBar->SetPct(1);                         // start meter
	OnPrintHeader(pDC,pInfo,NULL);	
	while (vpos != NULL)
		{
		pview = (CDibView* )pviewList->GetNext(vpos);  
		i++;
		if (i < pgStart+1)	continue;                 // ignore views before this page
		tr.SetRectEmpty();                            // empty the target rectangle
		tr.left=pInfo->m_rectDraw.left+c*(cxPrint/grid.x); // view will render self into tr
		tr.right=tr.left + w; 
		tr.top= pInfo->m_rectDraw.top+ r*(cyPrint/grid.y);
		tr.bottom=tr.top + h; 
		pview->OnPrintCaption(pDC, pInfo,&tr,NULL);        //print caption
		pview->OnPrintThumb(pDC, pInfo,tr);                //print thumbnail 
		if (!pInfo->m_bPreview)
			pStatusBar->SetPct(i * 100 / uCnt);                //update status bar meter
		c++; 
		if (!( c < WORD(grid.x)))
			{                                                // move down a row, reset column
			c=0;
			r++;
			}  
		if (i > pgEnd)                                     // used by OnPrepareDC()
			{
			m_bIsMoreToPrint = TRUE;
			break;
			}
		} 
	if (!pInfo->m_bPreview)                
		pStatusBar->SetPct(0);                               // turn off meter
}


void CDibView::OnPrintThumb(CDC* pDC,CPrintInfo* pInfo, CRect tr) 
{                    //tr is nominal target rectangle (before image scaling)
	CDibDoc* 			pDoc = GetDocument(); 
	CMainFrame* 	pMF = GetMainFrame();	    
	CMDIChildWnd* pDocFrame = GetDocFrame();
	HDIB 					hDIB = pDoc->GetHDIB();
	WORD					cx,cy,maxxy;
	
	if (hDIB != NULL)
		{                //get DIB parameters
		LPSTR lpDIB = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);
		int cxDIB = (int) ::DIBWidth(lpDIB);         // Size of DIB - x
		int cyDIB = (int) ::DIBHeight(lpDIB);        // Size of DIB - y
		::GlobalUnlock((HGLOBAL) hDIB);
		CRect rcImg;     //establish image rect
		rcImg.top = rcImg.left = 0;
		rcImg.right = cxDIB;
		rcImg.bottom = cyDIB;
		CRect rcDest;    //comp scaled DIB destination rect
		maxxy=max(cxDIB,cyDIB); 
		cx=tr.Width();
		cy=tr.Height();
		if (cx < cy)
			tr.bottom = tr.top + cx;
		else
			tr.right = tr.left + cy; 
		cx=tr.Width();
		cy=tr.Height();
		tr.right=tr.left+ ::MulDiv(cx,cxDIB,maxxy);
		tr.bottom=tr.top+ ::MulDiv(cy,cyDIB,maxxy);
		
		::PaintDIB(pDC->m_hDC, &tr, pDoc->GetHDIB(),
					&rcImg, pDoc->GetDocPalette());
		}
}

void CDibView::OnPrintCaption(CDC* pDC,CPrintInfo* pInfo, CRect* ptr,CString* pfile)
{
  CString caption ;
  
  GetParentFrame()->GetWindowText(caption);
  
	CSize dtSize = pDC->GetOutputTextExtent(caption,caption.GetLength()); 
	
	CRect dtRect(*ptr); 
	dtRect.top = dtRect.bottom - dtSize.cy;
	
	pDC->DrawText(caption,-1,dtRect,DT_LEFT);	
  ptr->SubtractRect(ptr,&dtRect);
}

void CDibView::OnPrintHeader(CDC* pDC,CPrintInfo* pInfo,CString* pfile)
{   	
	char buf[80];        
	ostrstream msg(buf,sizeof(buf));
  CString caption,caption1 ;

  msg<<"Otto's Album -  page:"<<pInfo->m_nCurPage <<ends;
  caption1 = buf;  
  
	CSize dtSize = pDC->GetOutputTextExtent(caption1,caption1.GetLength()); 
	
	CRect dtRect(pInfo->m_rectDraw); 
	dtRect.bottom = dtRect.top + dtSize.cy;
	
	pDC->DrawText(caption1,-1,dtRect,DT_LEFT);	
  pInfo->m_rectDraw.top += dtSize.cy * 2;    //adjust the drawing rect
} 

void CDibView::OnSize(UINT nType, int cx, int cy)
{
	CScrollView::OnSize(nType, cx, cy);
  if (GetParentFrame()->IsZoomed())
		SetScrollSizes(MM_TEXT,GetDocument()->GetDocSize()); 
	else                     
		SetScrollSizes(MM_TEXT,GetDocument()->GetThumb()->Size());
	ScrollToPosition(CPoint(0,0));  
}

///////////////////////  On Update Handlers ////////////////////////////
void CDibView::OnUpdateEditCopy(CCmdUI* pCmdUI)
{
	pCmdUI->Enable(GetDocument()->GetHDIB() != NULL);
}


void CDibView::OnUpdateEditPaste(CCmdUI* pCmdUI)
{
	pCmdUI->Enable(::IsClipboardFormatAvailable(CF_DIB) ||
		::IsClipboardFormatAvailable(CF_METAFILEPICT));
}


/////////////////////////  helper function  ////////////////////////////
void Rectangle3D(CDC* pDC, CRect r, UINT style )
{
	CPen* 			poldPen;
	CBrush			ltgrayBrush; 
	CPen				whitePen, grayPen;
		
		ltgrayBrush.CreateStockObject(LTGRAY_BRUSH);
		pDC->FillRect(&r,&ltgrayBrush);
		
		whitePen.CreateStockObject(WHITE_PEN);
		grayPen.CreatePen(PS_SOLID,1,0x00808080);
		
		poldPen = pDC->SelectObject(&whitePen);
		pDC->MoveTo(r.left, r.bottom);
		pDC->LineTo(r.left, r.top);
		pDC->LineTo(r.right, r.top);
		pDC->SelectObject(&grayPen);
		pDC->LineTo(r.right,r.bottom);
		pDC->LineTo(r.left, r.bottom);
		pDC->SelectObject(poldPen);

} 


