{$F+,A+,R-,S-,V-,O-,G+}
{GIF DLL D Overmyer} {tabs = 2}
library GIF;             {basic GIF image decoder}
uses WinTypes,WinProcs,Objects,OMemory,OWindows,Win31,Graphics;
Const
	LargestCode = 4095;
  Mask: Array[0..11] of Integer=
			($0001,$0003,$0007,$000F,
       $001F,$003F,$007F,$00FF,
       $01FF,$03FF,$07FF,$0FFF);
	BuffSize = 8192*2;
	st_OK = 1;
  st_ClipFailure = 3;
  st_NoMem = 4;
  st_Fail = 2;
type
	BuffType = Array[0..BuffSize] of Byte;
  BuffPtr = ^BuffType;

  GetByteProc = function : Byte;
  PutByteProc = function(Pix:Integer):Boolean;
  MapType = (Global, Local);
  GifBlockType = Array[0..255] of Byte;
  Primary = (RedVal,GreenVal,BlueVal);
  MapEntry = array[RedVal..BlueVal] of Byte;

	GIFMap = record
    Map              : array[0..255] of MapEntry;
    MapExists        : Boolean;
    Sorted           : Boolean;
    BitsPerPixel     : Word;
    HighColorNum     : Word;
    IsGlobal         : Boolean;             {only true  if Global}
    BackgrColorIndex : Word;                {only valid if Global}
    AspectRatio      : Word;                {only valid if Global}
    BitsPerPrimary   : Word;                {only valid if Global}
  	Interlaced       : Boolean;             {only valid if Local }
	end;
	CodeEntry =Record
  	Prefix: Integer;   { 2 bytes }
  	Suffix: Byte;      { 1 byte  }
  	Stack:  Byte;      { 1 byte  }
		end;                 { 4096 * 4 = 16k }
	CodeTableType = Array[0..LargestCode] of CodeEntry;
	TablePointer  = ^CodeTableType;

var
	ReadBuff  :BuffPtr;
  hReadBuff :THandle;
  WriteBuff :BuffPtr;
  hWriteBuff:THandle;
  GIFBits	  :Pointer;
  hGIFBits	:THandle;
  pGIFBits  :Pointer;
  Actual		:Word;
  ReadIdx   :Word;
  WriteIdx  :Word;
  ReadCnt   :Word;
  WriteCnt  :Word;
	F1				:THandle;
  cDIBLine 	:Word;
  cGIFLine 	:Word;

  GetByte      : GetByteProc;
  PutByte      : PutByteProc;
  ExtendFunc : Byte;         {Function code for extension block}
  GIFSig     : String[6];    {GIF ID string usually = 'GIF87a'}
  ImageLeft,                 {Left edge of image relative to virtual screen}
  ImageTop,                  {Top edge of image relative to virtual screen}
  ImageWidth,                {in pixels}
  ImageHeight,               {in pixels}
  LeftEdge,
  RightEdge,
  ScrColors,
  ScrHeight,                 {in pixels}
  ScrWidth   : Word;         {in pixels}
  PackedBits, I : Word;
  A, B : Byte;
  BytesInBlock : Byte;
  Maps        : Array[MapType] of GIFMap;
  CurMap      : MapType;
  TempMap     : GIFMap;
  Color       : array[0..255] of byte;
  MaxColors   : Integer;
  IsInterlaced:Bool;

	CodeSize,        ClearCode,
  EOFCode,         FirstFree,
  BitOffset,       BytOffset,
  BitsLeft,        MaxCode,
  FreeCode,        OldCode,
  InputCode,       Code,
  SuffixChar,      FinalChar,
  MinimumCodeSize, BytesUnRead:     Integer;
  CodeBuffer: Array[0..63] of Byte;
  CodeTable: TablePointer;
  ExpError: Integer;
{************************* Functions  *******************************}
function LongMin(A, B: LongInt): LongInt;
begin
  if A < B then LongMin := A else LongMin := B;
end;

function LongMax(A, B: LongInt): LongInt;
begin
  if A > B then LongMax := A else LongMax := B;
end;

function FileGetByte:Byte;far;
begin
	if ReadIdx >= ReadCnt then begin
  	ReadCnt:=_lRead(F1,PChar(ReadBuff),BuffSize);
    ReadIdx :=0;
  end;
  FileGetByte := ReadBuff^[ReadIdx];
  Inc(ReadIdx);
end;

function BufPutByte(Pix:Integer):Boolean;
begin
	if Pix = MaxInt then
  	begin
    hMemCpy(pGIFBits,WriteBuff,WriteIdx);
  	BufPutByte := False;
    Exit;
    end
  else
  	begin
		if WriteIdx >= BuffSize then
			begin
			hMemCpy(pGIFBits,WriteBuff,BuffSize);
    	IncPtr(pGIFBits,BuffSize);
    	WriteIdx :=0 ;
  		end;
  	WriteBuff^[WriteIdx] := Pix;
  	Inc(WriteIdx);
  	BufPutByte:= True;
  	end;
end;

function GetWord : word;
begin
	a := GetByte;
  b := GetByte;
  GetWord := (b shl 8) or a;
end;

procedure GetBlock(var Block : GifBlockType);
begin
	Block[0] := GetByte;
  if Block[0] <> 0 then
  	for I := 1 to Block[0] do Block[I] := GetByte;
end;

procedure GetGIFSig;
var I : Integer;
begin
  GIFSig := '';
  for I := 0 to 5 do
    GIFSig := GIFSig + chr(GetByte);
end;

procedure GetScreenDesciptor(var MapRec : GifMap);{-get a screen descriptor record}
begin
  ScrWidth := GetWord;
  ScrHeight := GetWord;
  PackedBits := GetByte;
  IsInterlaced := False;
  with MapRec do begin
    IsGlobal := true;
    Interlaced := false; {undefined}
    BitsPerPrimary := ((PackedBits and $70) shr 4) + 1;
    BackgrColorIndex := GetByte;
    MapExists := (PackedBits and $80) <> 0;
    BitsPerPixel := (PackedBits and $7) + 1;
    HighColorNum := (1 shl BitsPerPixel)-1;
    ScrColors := Succ(HighColorNum);
    Sorted := (PackedBits and $04) <> 0;
    AspectRatio := GetByte;
    if MapExists then  {get the map}
     for I := 0 to HighColorNum do begin
       Map[I,RedVal] := GetByte;
       Map[I,GreenVal] := GetByte;
       Map[I,BlueVal] := GetByte
     end;
  end;
end;

procedure GetImageDescriptor(var MapRec : GifMap); {-get an image descriptor record}
begin
  ImageLeft := GetWord;
  ImageTop := GetWord;
  ImageWidth := GetWord;
  ImageHeight := GetWord;
  PackedBits := GetByte;
  IsInterlaced := (PackedBits and $40) <> 0;
  with MapRec do begin
    IsGlobal := false;
    AspectRatio := 0;      {undefined}
    BitsPerPrimary := 0;   {undefined}
    BackgrColorIndex := 0; {undefined}
    Interlaced := (PackedBits and $40) <> 0;
    Sorted := (PackedBits and $20) <> 0;
    MapExists := (PackedBits and $80) <> 0;
    BitsPerPixel := (PackedBits and $7)+1;
    HighColorNum := (1 shl BitsPerPixel)-1;
    if MapExists then
     for I := 0 to HighColorNum do begin
       Map[I,RedVal] := GetByte;
       Map[I,GreenVal] := GetByte;
       Map[I,BlueVal] := GetByte
     end;
  end;
end;

function GetExtendFunc : Byte;
begin
  GetExtendFunc := GetByte;
end;

function GetExtendBlock(var Block : GifBlockType) : Boolean;
begin
  GetBlock(Block);
  GetExtendBlock := (Block[0] <> 0);
end;

procedure SkipExtendBlock; {-skip 89a-spec extension block}
var
  Block : GifBlockType;
begin
  GetExtendFunc;
  while GetExtendBlock(Block) do ;
end;

Procedure InitializeTable(MinCodeSize: Byte);
begin
  CodeSize:=  Succ(MinCodeSize);
  ClearCode:= 1 Shl MinCodeSize;
  EOFCode:=   Succ(ClearCode);
  FirstFree:= Succ(EOFCode);
  FreeCode:=  FirstFree;
  MaxCode:=   1 Shl CodeSize;
end;

{  ReadCode needs to return an integer made up of a variable number of bits
  (actually 3..12) from the input data stream. Since 12 bits may at any time
  span THREE input bytes, the routine was designed using a LONG INTEGER
  (32 bit) type.
  For example: Looking for 10 bits from the stream (B's):

               Byte 3     Byte 2     Byte 1
              ---------  ---------  ---------
              cccc cccB  BBBB BBBB  Baaa aaaa
                      -  ---- ----  -
     Three bytes are loaded into a Longint record,
     they are shifted right by the number in BitsLeft,
     AND'ed with Mask[CodeSize-1],
     and returned (as an integer) as ReadCode}
function ReadCode: Integer;
type
 	CodeRec = record
 		case Integer of
  		0:(L:LongInt);
    	1:(W1:Word;
    	 	 W2:Word);
    	2:(B1:Byte;
    	 	 B2:Byte;
       	 B3:Byte;
       	 B4:Byte);
 end;
var
	I, Ch: Integer;
  BufPointer: Byte;
  RC: Integer;
  Reading: Boolean;
  CX:CodeRec;
begin
  BytOffset:= BitOffset Shr 3;
  BitsLeft:=  BitOffset And 7;
  If BytOffset>=61 then
    begin
    BufPointer:=0;
    For I:=BytOffset to 63 do
      begin
      CodeBuffer[BufPointer]:=CodeBuffer[I];
      BufPointer:=Succ(BufPointer);
      end;
    Reading:=True;
    While Reading do
      begin
      If BytesUnRead=0 then BytesUnRead:=GetByte;
      If BytesUnRead<1 then
       begin
       Reading:=False;
       If BytesUnRead<0 then ExpError:=BytesUnRead;
       end;
      If Reading then
       begin
       Ch:=GetByte;
       CodeBuffer[BufPointer]:=Ch;
       BytesUnRead:=Pred(BytesUnRead);
       BufPointer:=Succ(BufPointer);
       Reading:=(BufPointer<64);
       end;
      end;
    BitOffset:=BitsLeft; BytOffset:=0;
    end;
    BitOffset:=BitOffset+CodeSize;
    CX.L:=0;
    CX.B1:=CodeBuffer[BytOffset];
		CX.B2:=CodeBuffer[BytOffset+1];
		CX.B3:=CodeBuffer[BytOffset+2];
    CX.W1 := Word((CX.L shr BitsLeft)AND Mask[CodeSize-1]);
    ReadCode :=CX.W1;
end;

function ExpandGif: Integer;   {expand the input code data}
 Var
 	I, SP:         Integer;
  MemAvail:      Real;
  MAvail:        Integer;
begin
  New(CodeTable);
  FillChar(CodeTable^,SizeOf(CodeTableType),0);
  ExpError:=0;

  Code:=0; OldCode:=0; SuffixChar:=0; FinalChar:=0;
  MinimumCodeSize:=GetByte;

  If MinimumCodeSize < 0 then
		ExpError:=MinimumCodeSize
  else
		If not (MinimumCodeSize in [2..9]) then
			ExpError:=-3
    else
      begin
      InitializeTable(MinimumCodeSize);
      Sp:=0; BytesUnRead:=0; BitOffset:=64*8;       { force first-time read }

      Repeat { Until (Code=EOFCode) or (ExpError<>0) }
        Code:=ReadCode;
        If Code=EOFCode   then
					begin
					end        { reached end of image }
         else
				 	 If Code=ClearCode then                { table full, start over }
            begin
            InitializeTable(MinimumCodeSize);
            Code:=       ReadCode;
            OldCode:=    Code;
            SuffixChar:= Code;
            FinalChar:=  Code;
            If not PutByte(SuffixChar) then
              begin
							ExpandGif:=-1; Exit;
							end;
            end
          else
            begin
            InputCode:=Code;
            If Code >= FreeCode then
              begin
              Code:= OldCode;
              CodeTable^[SP].Stack:= FinalChar;
              SP:= Succ(SP);
              end;
            While Code >= FirstFree do
              begin
              CodeTable^[SP].Stack:= CodeTable^[Code].Suffix;
              Code:= CodeTable^[Code].Prefix;
              SP:= Succ(SP);
              end;
            FinalChar:=  Code;
            SuffixChar:= Code;
            CodeTable^[SP].Stack:= FinalChar;
            SP:= Succ(SP);
            While (SP>0) do
              begin
              SP:=Pred(SP);
              If not PutByte(CodeTable^[SP].Stack) then
                 begin
								 ExpandGif:=-1;
								 Exit;
								 end;
              end;
            CodeTable^[FreeCode].Suffix:= SuffixChar;
            CodeTable^[FreeCode].Prefix:= OldCode;
            FreeCode:= Succ(FreeCode);
            OldCode:= InputCode;
            If FreeCode >= MaxCode then
              If CodeSize < 12 then
                begin
                CodeSize:= Succ(CodeSize);
                MaxCode:=  MaxCode Shl 1;
                end;
            end;
        Until (Code=EOFCode) or (ExpError<>0);
      end;
  PutByte(MaxInt);
  Dispose(CodeTable);
  ExpandGif:=ExpError;
end;   { Function ExpandGif }

function GIFtoDIB(fName:PChar):THandle; export;
const
  YInc :Array[1..6] of Byte=(8,8,4,2,1,0);
  Ylin :Array[1..6] of Byte=(0,4,2,1,0,0);
var
	IsDone:Bool;
  BlockType:Char;
  Blk:GifBlockType;
  ExtFunc:Byte;
  Actual   	:Word;
  FS			 	:LongInt;      {file size}
  GIFLine  	:PByteArray;   {Buffer to hold 1 expanded GIF bits scanline}
  DIBLine  	:PByteArray;   {Buffer to hold 1 DIB scan line after repacking}
  Res				:Integer;      {resolution of image in bits per pixel}
	cColor		:Integer;      {color count of color map (palette)}
  cDIBBits  :LongInt;      {size in bytes of DIB image bits}
  cGIFBits  :LongInt;      {size in bytes of GIF bits after expansion}
  hDIB      :THandle;      {handle for packed DIB memory block}
  bi				:PBitmapInfo;  {pointer to packed DIB = locked hDIB}
  DIBBits   :Pointer;      {pointer to memory block receiving DIB bits}
  pDIBBits	:Pointer;      {pointer to memory block receiving DIB bits}
  I,J				:Integer;
  F					:THandle;

  {repacking the bits involves 3 problems:
  	1. one bitperpixel images use only first bit of each byte in scanline
       four bpp images use only first nibble of each byte in scanline
       eight bpp images use entire byte of scanline
    Thus the input bytes must be repackaged appropriately.

    2. Interlaced images have scanlines which are non-sequential.
			They are ordered as follows:
    		pass 1, every eighth scanline beginning with line 0
        pass 2, every eighth scanline beginning with line 4
        pass 3, every fourth scanline beginning with line 2
        pass 4, every second scanline beginning with line 1
    Thus, these interlaced scanlines must be resequenced as needed.

    3. Packed DIB images are stored inverted in memory,
			i.e. last scanline is first, first is last.
    Therefore the GIF scanlines are resequenced by writing to the end of the
    DIB bits block, and backing up the bits pointer as needed}

	function RepackBits:Boolean ;
	var
	I,J,Pass,Row:Integer;
	begin
	pGIFBits := GIFBits;             {start with first GIF scan line}
	Pass := 1; Row:= 0;
  for J := 1 to ImageHeight do     {one scanline at a time}
  	begin
    FillChar(DIBLine^,cDIBLine,0);
    hMemCpy(GIFLine,pGIFBits,cGIFLine);  {get a scanline to work on}
    IncPtr(pGIFBits,cGIFLine);           {increment pointer to GIF bits}
  	if Res = 1 then                      {if monochrome}
    	begin                              {use only low-order bit from each byte}
      for i := 0 to ((cGIFLine-1) div 8) do
    		DIBLine^[i] :=                   {form a byte from the 8 bits}
					((GIFLine^[(8*I)+0] AND $01) shl 7) OR
					((GIFLine^[(8*I)+1] AND $01) shl 6) OR
					((GIFLine^[(8*I)+2] AND $01) shl 5) OR
					((GIFLine^[(8*I)+3] AND $01) shl 4) OR
 					((GIFLine^[(8*I)+4] AND $01) shl 3) OR
					((GIFLine^[(8*I)+5] AND $01) shl 2) OR
					((GIFLine^[(8*I)+6] AND $01) shl 1) OR
					((GIFLine^[(8*I)+7] AND $01) shl 0);
    	end
		else if Res = 4 then                 {if 16 color bitmap}
    	begin                              {use only low-order 4 bits}
    	for i := 0 to ((cGIFLine-1) div 2) do  {combine nibbles to make a byte}
    		DIBLine^[i] :=((GIFLine^[(2*I)+1] AND $0F)) OR
									  ((GIFLine^[(2*I)+0] AND $0F) shl 4);
    	end
		else if Res = 8 then                 { if 256 color }
      hMemCpy(DIBLine,GIFLine,cGIFLine); {just copy the bytes to DIB scanline}

    if IsInterlaced then                 {if interlaced}
    	begin                              {copy DIB scanline to pointer}
   		hMemCpy(pDIBBits,DIBLine,cDIBLine);
      Row:= Row+YInc[Pass];              {calculate position of next scanline}
      if (Row) > ImageHeight-1 then      {if this pass is finished, calculate next}
				begin
				Inc(Pass);                       {up the pass count}
        Row:=YLin[Pass];                 {get the starting row from array}
        pDIBBits := DIBBits;             {point to last scanline in DIBBits}
        IncPtr(pDIBBits,-LongMul(cDIBLine,Row)); {move backwards based on row count}
        end                              {pDIBBits ready now for next scanline}
      else
        IncPtr(pDIBBits,-LongMul(cDIBLine,YInc[Pass])); {move backwards in DIBBits}
      end
    else                                  {No interlace}
    	begin
   		hMemCpy(pDIBBits,DIBLine,cDIBLine); {just copy the scanline to DIBBits}
   		IncPtr(pDIBBits,-cDIBLine);         {backup the pointer}
      end;
	 end;
	end;

	function SetupDIB:Boolean;              {allocate memroy for packed DIB image block}
  var
  	I:Integer;
	begin
  SetupDIB:= False;
  cDIBBits :=GetDIBSize(ImageWidth,ImageHeight,Res);  {get size of DIB bits array}
  cDIBLine := cDIBBits div (ImageHeight);          {size of DIB scan line}
  cColor:=Maps[Global].HighColorNum+1;             {size of color map/palette}
  hDIB :=GlobalAlloc(GHND,
				sizeof(TBitmapInfoHeader)+(cColor*sizeof(TRGBQuad))+cDIBBits);
  if hDIB = 0 then Exit;
  bi := GlobalLock(hDIB);
  with bi^.bmiHeader do                           {setup DIB header}
  	begin
		biSize:= sizeof(TBitmapInfoHeader);
  	biWidth :=ImageWidth;
  	biHeight :=ImageHeight;
  	biPlanes := 1;
  	biBitCount := RES;
  	biCompression := BI_RGB;
  	biSizeImage:=cDIBBits;
		end;
  if Maps[Local].MapExists then
   	CurMap := Local
	else
		CurMap := Global;
  with Maps[CurMap] do                          {xfer color map to DIB}
  	for i := 0 to  (HighColorNum) do
  		begin
  		bi^.bmiColors[i].RGBblue := Map[i,BlueVal];
  		bi^.bmicolors[i].RGBred  := Map[i,RedVal];
  		bi^.bmicolors[i].RGBGreen := Map[i,GreenVal];
  		bi^.bmiColors[i].RGBReserved := 0;
  		end;
  pDIBBits:=GetPtr(bi,sizeof(TBitmapInfoHeader)+cColor*sizeof(TRGBQuad));
 	IncPtr(pDIBBits,cDIBBits-cDIBLine);   			 {offset bits ptr to last DIB scanline}
  DIBBits := pDIBBits;                         {used to decode interlaced images}
  DIBLine := GlobalAllocPtr(GHND,cDIBLine);    {get buffer for DIB scan line}
  if DIBLine = nil then                        {if fail, deallocate buffers}
  	GlobalFreePtr(bi)
  else
  	SetupDIB:=True;
	end;

	function SetupGIF:Boolean;                  {setup GIF buffers}
	begin
  GIFLine := nil;
	SetupGIF:=False;
  cGIFBits := LongMul(ImageWidth,ImageHeight); {get size of GIF bits in bytes}
  cGIFLine := ImageWidth;                      {size of GIFLine buffer in bytes}
	GIFLine := GlobalAllocPtr(GHND,cGIFLine);    {alloc GIF line buffer}
  hGIFBits:=GlobalAlloc(GHND,cGIFBits);        {alloc GIFBits block to hold uncompressed bits}
  GIFBits :=GlobalLock(hGIFBits);
  pGIFBits:=GIFBits;                           {point to top of GIFBits}
  if GIFBits = nil then
  	begin                   									 {unring the bell}
		GlobalFreePtr(bi);
		GlobalFreePtr(DIBLine);
		hDIB:=0;
    if GIFLine <> nil  then
  		GlobalFreePtr(GIFLine);
    end
  else
  	SetupGIF:=True;
	end;

  function FreeBuffers:Bool;                  {deallocate buffers}
  begin
  FreeBuffers:=True;
  GlobalFreePtr(GIFBITS);
  GlobalFreePtr(GIFLine);
  GlobalFreePtr(DIBLine);
  end;

begin                                 {assume failure}
	GIFtoDIB:=0;hDIB:=0;
  F:=_lopen(fName,of_Read);
  if F = -1 then Exit;
	F1:=F;IsDone:=False;IsInterlaced := False;
  ReadCnt:=0;ReadIdx:=0;WriteIdx:=0;
	ReadBuff:=GlobalAllocPtr(GHND,BuffSize);    {for buffered read/writes}
  WriteBuff:=GlobalAllocPtr(GHND,BuffSize);
  GetByte := FileGetByte;  {assign the function pointers}
  PutByte:=BufPutByte;
  ReadIdx:=MaxInt;         {force a read fault in GetByte()}
  CurMap:=Global;          {ready for ScreenDescriptor}

  GetGIFSig;               {validate the input stream}
  GetScreenDesciptor(Maps[CurMap]); {get screen Descriptor}
  Res:= Maps[CurMap].BitsPerPixel;

  while not IsDone do
  	begin
    BlockType :=Chr(GetByte);
    case BlockType of
    	',':                      {start of image descriptor}
      	begin
        GetImageDescriptor(Maps[Local]);
        CurMap:= Global;
  			if Res = 0 then Res:= Maps[Local].BitsPerPixel;     {resolution is global}
        if not SetupDIB then Break;          {setup buffers}
        if not SetupGIF then Break;
        if ExpandGif = 0 then                {expand the GIF blocks}
        	begin
        	RepackBits;                        {repack the bits into DIB format}
					GlobalUnlock(hDIB);                {we're done}
          end
        else
        	begin                              {freeup hDIB memory}
					GlobalFreePtr(bi);
					hDIB:=0;
					end;
        FreeBuffers;                         {deallocate DIB/GIF buffers}
        IsDone:= True;                       {we do one image only}
        end;
      '!':                  {extension block}
      	begin
        while GetExtendBlock(Blk) do;
        end;
      ';':
      	begin
        IsDone:=True;
        end;
    end;
  end;
  GlobalFreePtr(ReadBuff);                   {free read/write buffers}
  GlobalFreePtr(WriteBuff);
  GIFtoDIB:=hDIB;
  _lclose(F);
end;

exports
	GIFtoDIB index 1;

begin
end.
