// pcx.cpp define pcx functions
// doug overmyer 6/10/93

#include  	"stdafx.h"    
#include 		"dibapi.h"
#include  	"pcx.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif     

struct RGBT
{
	BYTE R,G,B ;
};

struct PCXHEADER
{
	BYTE flag,version,encoding,bpp;
	WORD xmin,ymin,xmax,ymax,hres,vres;
	RGBT rgb[16];
	BYTE reserved,planes;
	WORD bpl,paltype,videox,videoy;
	BYTE filler[54];
};

#define MAXBUF    (0x7FFF)
#define OVERRUN 	100

// Private functions
BOOL GetPCXLine(CBufFile* f, LPSTR PCXLine, WORD cPCXLine, WORD cMaxLine);

BOOL RepackBits(LPSTR PCXLine, LPSTR DIBLine, WORD cPCXLine,
								WORD cDIBLine, WORD  planes, WORD bpl,
								WORD bpp, WORD res);    

/////////////////////////  CBufFile /////////////////////////////////////
// this hack adapts CFile to provide a buffered read of binary data 
// output via GetByte()
// not all CFile functions supported!
IMPLEMENT_DYNAMIC(CBufFile,CFile)       

CBufFile::CBufFile():CFile()
{                 //alloc a global buffer to hold byte stream
	m_OfsStart=0;
	m_hBuf = ::GlobalAlloc(GHND,MAXBUF);
	m_Buf = (LPSTR)::GlobalLock(m_hBuf);
	m_Indx = MAXBUF;
}

CBufFile::CBufFile(int hFile):CFile(hFile)
{ 
	m_OfsStart=0;
	m_hBuf = ::GlobalAlloc(GHND,MAXBUF);
	m_Buf = (LPSTR)::GlobalLock(m_hBuf);
	m_Indx = MAXBUF;

}

CBufFile::CBufFile(const char* pszFileName, UINT nOpenFlags,DWORD OfsStart):
		CFile(pszFileName,nOpenFlags)
{
	m_OfsStart=OfsStart;
	m_hBuf = ::GlobalAlloc(GHND,MAXBUF);
	m_Buf = (LPSTR)::GlobalLock(m_hBuf);
	m_Indx = MAXBUF;
}

           
CBufFile::~CBufFile()
{                      //free buffer
	::GlobalUnlock(m_hBuf);
	::GlobalFree(m_hBuf);
	CFile::~CFile();	
}
           
BYTE CBufFile::GetByte()
{    //check for buffer fault;if needed, read buffer;return next bye
	DWORD actual;
	if( m_Indx > (MAXBUF-1) )
		{
		_fmemset(m_Buf,'\x00',MAXBUF);
		actual = Read(m_Buf,MAXBUF);
		m_Indx = 0;
		}
	return (BYTE)m_Buf[m_Indx++];
}
           
           
LONG CBufFile::Seek(LONG lOff,UINT nFrom)
{    //clear buffer,set m_Index to fault on GetByte, seek in file
	_fmemset(m_Buf,'\x00',MAXBUF);
	m_Indx=MAXBUF;
	return CFile::Seek(lOff,nFrom);
}

////////////////////   GetPCXLine()  /////////////////////////
// private function to read f from current position, expand compressed bytes,
//   and return a complete scan line
BOOL GetPCXLine(CBufFile* f, LPSTR PCXLine, WORD cPCXLine, WORD cMaxLine)
{
	LPSTR			pPCXLine;
	WORD			tPCXLine = 0;
	BYTE			B;
	WORD			cB;
	
	_fmemset(PCXLine,'\x00',cMaxLine);
	pPCXLine = PCXLine;
	while (tPCXLine < cPCXLine)
		{
		cB = 1;
		B= f->GetByte();
		if ( 0xC0 == (0xC0 & B) )
			{
			cB = 0x3F & B;
			B = f->GetByte();
			}
		for (WORD j=1;j < min(cB,(cMaxLine-tPCXLine))+1;j++)
			{
			*pPCXLine = char(B);
			pPCXLine +=1;
			}
		tPCXLine +=cB;
		}
	return TRUE;
}


///////////////////////////  RepackBits() ///////////////////////////
//private function to repack pcx scan line to DIB scan line
//DIB resolution is match to incoming pcx file
//PCX format based on res

BOOL RepackBits(LPSTR PCXLine, LPSTR DIBLine, WORD cPCXLine,
								WORD cDIBLine, WORD  planes, WORD bpl,
								WORD bpp, WORD res)
{
	BYTE 		A0,A1,A2,A3;
	BYTE 		C;
	WORD		j = 0;
	
	if ( (res==1) || (res==8) )   // same format, just copy buffers
		hmemcpy( (void huge*)DIBLine, (void huge*)PCXLine,cDIBLine );
	else if (res == 4)      // must repack the bits from the 4 pcx color planes
		{                     // into 1 DIB scan line.
		for ( WORD i=0; (i<bpl);i++ ) 
			{
			A0=BYTE(PCXLine[i]);
			A1=BYTE(PCXLine[i+bpl]);
			A2=BYTE(PCXLine[i+(2*bpl)]);
			A3=BYTE(PCXLine[i+(3*bpl)]);
			
			for ( WORD k=0; (k<4);k++ )
				{
				C=((A0 & 0x80) >> 3) | ((A1 & 0x80) >> 2) |
		      ((A2 & 0x80) >> 1) | ((A3 & 0x80) >> 0) |  
		      ((A0 & 0x40) >> 6) | ((A1 & 0x40) >> 5) |
		      ((A2 & 0x40) >> 4) | ((A3 & 0x40) >> 3) ;
		    DIBLine[j]=char(C);
		    j++;
		    A0=A0 << 2;
		    A1=A1 << 2;
		    A2=A2 << 2;
		    A3=A3 << 2;
		    }
		  }
		} 
	else if (res == 24)  // take 1 byte from 3 color planes, move to output buffer
		{   
		for (WORD i=0;(i<bpl);i++)
		  {
		  DIBLine[j+2]=PCXLine[i];
		  DIBLine[j+1]=PCXLine[i+bpl];
		  DIBLine[j+0]=PCXLine[i+(2*bpl)];
		  j +=3;
		  }
		} 
	return TRUE;
}	  
                                              
///////////////////////////  LoadPCXFile()
HDIB LoadPCXFile(PSTR fname, DWORD ofsstart)
{
	HDIB 				hDIB;
	PCXHEADER		pcxh;
	DWORD				fsize, cDIBBits =0;
	LPSTR				PCXLine, DIBLine;
	HGLOBAL     hPCXLine, hDIBLine;
	WORD				cPCXLine=0, cDIBLine=0, cMaxLine =0;
	WORD				res=0, cColor=0;
//	HGLOBAL			hbi;
	BITMAPINFO far *	bi;
	char huge *	pBits;
	RGBT				rgb;
	WORD				xSize=0, ySize=0;
	CBufFile		f;    
//	CFile       f;
	CFileException fe;
	DWORD				actual;  
	LONG				fpos;
	
	if ( !f.Open(fname, CFile::modeRead | CFile::shareDenyWrite, &fe)  )
		return 0;

	fsize=f.GetLength();
	f.Seek(ofsstart,CFile::begin);
	actual=f.Read(&pcxh,sizeof(pcxh));
	if (pcxh.version != 5)
		return 0;
	res=pcxh.planes * pcxh.bpp;
	if ( !( (res == 1) | (res==4) | (res==8) | (res==24) )  )
		return 0;   
		
	xSize=pcxh.xmax - pcxh.xmin +1;
	ySize=pcxh.ymax - pcxh.ymin +1;
	
	if (res < 24) cColor = 1 << res;
	cDIBBits = ((DWORD(xSize * res) + 31) / 32 * 4) * DWORD(ySize);
	
	cPCXLine = pcxh.planes * pcxh.bpl;
	cDIBLine = (WORD) (cDIBBits / DWORD(ySize));
	cMaxLine = max(cPCXLine, cDIBLine);
	
	hDIB = (HDIB)::GlobalAlloc(GHND,sizeof(BITMAPINFOHEADER) +
									(cColor*sizeof(RGBQUAD)) + cDIBBits);
	if (hDIB == 0) return 0;
	
	bi = (BITMAPINFO far *)::GlobalLock(HGLOBAL(hDIB));
	if (bi == NULL)
		{
		::GlobalFree(hDIB);
		return 0;
		}
		
	bi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi->bmiHeader.biWidth = xSize;
	bi->bmiHeader.biHeight = ySize;
	bi->bmiHeader.biPlanes = 1;
	bi->bmiHeader.biBitCount = res;
	bi->bmiHeader.biCompression = BI_RGB;
	bi->bmiHeader.biSizeImage = cDIBBits;
	
	pBits = (char huge *) bi;
	pBits +=sizeof(BITMAPINFOHEADER) + (cColor*sizeof(RGBQUAD));
			
	if (res < 5) 
		fpos=f.Seek(16,CFile::begin);
	else if (res < 9)
		fpos=f.Seek(-768,CFile::end);
	
	if (res < 24)
		for (WORD i=0;i<cColor;i++)
			{                                
			f.Read(&rgb,sizeof(RGBT));
			bi->bmiColors[i].rgbBlue = rgb.B;  
			bi->bmiColors[i].rgbRed  = rgb.R;		
			bi->bmiColors[i].rgbGreen= rgb.G;
			bi->bmiColors[i].rgbReserved = 0;	
			}
	
	f.Seek(sizeof(PCXHEADER),CFile::begin); 
	
	hPCXLine = (HGLOBAL)::GlobalAlloc(GHND,cMaxLine);             
	PCXLine = (LPSTR)::GlobalLock(hPCXLine);
	hDIBLine = (HGLOBAL)::GlobalAlloc(GHND,cMaxLine);
	DIBLine = (LPSTR)::GlobalLock(hDIBLine);
		
	pBits = pBits + (cDIBBits-cDIBLine);    //point to last dib scan line                                              
	
  for (WORD i = 0; i<ySize;i++)
  	{
  	GetPCXLine(&f,PCXLine, cPCXLine,cMaxLine);
  	RepackBits(PCXLine,DIBLine,cPCXLine,cDIBLine,pcxh.planes,pcxh.bpl,
  						 pcxh.bpp,res);
  	::hmemcpy(pBits,DIBLine,cDIBLine);
  	pBits = pBits - cDIBLine;
  	}
  
  ::GlobalUnlock(hPCXLine);
  ::GlobalFree(hPCXLine);
  ::GlobalUnlock(hDIBLine);
  ::GlobalFree(hDIBLine);
  
  ::GlobalUnlock(hDIB);
  
		
	return hDIB;
}	